/************************************************************************************
	Il seguente programma implementa il comando:
		gunzip -c %1 | grep -i %2 | more
	dove
		%1 e %2 sono due parametri passati su linea di comando (argv[1] e argv[2])

	+--------+   p1   +------+   p2    +------+
	| gunzip |=======>| grep |========>| more |
	+--------+  pipe  +------+  pipe   +------+
     processo         processo         processo

	Per realizzare il comando sono necessari:
		- 3 processi: per aggiungere un nuovo processo nel sistema si utilizza la
					chiamata di sistema fork()
		- 2 pipe: si usa la chiamata di sistema pipe() per creare una nuova pipe

	NOTA: le condizioni di errore non sono gestite. Bisogna sempre verificare il
		  codice di ritorno di tutte le chiamate di sistema che si usano. 
************************************************************************************/

#include <iostream>
#include <unistd.h>
#include <errno.h>

using namespace std;

int main(int argc, char* argv[])
{
	pid_t child;
	int p1[2];

	// Verifico che sulla linea di comando sono presenti i due parametri richiesti
	if(argc != 3)
	{
		cerr << "Uso: " << argv[0] << " nome_file.gz stringa" << endl;
		return 1;
	}

	// La linea di comando contiente tre elementi
	//		argv[0]	->	Nome del programma
	//		argv[1]	->	Primo parametro (%1)
	//		argv[2]	->	Secondo parametro (%2)


	// Creo la prima pipe. Bisogna creare la pipe prima di aggiungere un nuovo
	// processo nel sistema altrimenti il meccanismo non funziona.
	if(pipe(p1) == -1)
	{
		cerr << "pipe: " << errno << endl;
		return 2;
	}

	// Aggiungo un nuovo processo. Il nuovo processo erediter la pipe p1
	//  dal processo genitore.
	child = fork();

	if(child == -1)
	{
		cerr << "fork: " << errno << endl;
		return 3;
	}

	if(child == 0)
	{
		// Processo figlio

		// sostituisco lo standard output con il lato in scrittura della pipe
		int result = dup2(p1[1], 1); 
		// Da questo momento in poi lo standard output  la pipe

		close(p1[0]); close(p1[1]); // bisogna chiudere sempre i file che non servono
		
		if(result == -1)
		{
			cerr << "dup2: " << errno << endl;
			return 4;
		}
		
		// carico in memoria il programma gunzip. execlp NON aggiunge un nuovo
		// processo nel sistema: carica le istruzioni da eseguire dal file "gunzip".
		execlp("gunzip", "gunzip", "-c", argv[1], NULL);
		
		// il codice seguente  eseguito solo se execlp fallisce
		cerr << "exec(gunzip): " << errno << endl;
		return 5;
	}

	// processo genitore

	int p2[2];

	// creo una nuova pipe
	if(pipe(p2) == -1)
	{
		cerr << "pipe: " << errno << endl;
		return 6;		
	}
	
	// aggiungo un altro processo nel sistema
	child = fork();

	if(child == -1)
	{
		cerr << "fork: " << errno << endl;
		return 7;		
	}

	if(child == 0)
	{
		// secondo processo figlio

		// sostituisco lo standard input con il lato in lettura della prima pipe
		dup2(p1[0], 0);
		// da questo momento in poi lo standard input contiene i dati provenienti
		// dallo standard output del processo gunzip

		// sostituisco lo standard output con il lato in scrittura della seconda pipe
		dup2(p2[1], 1);
		
		close(p1[0]); close(p1[1]);
		close(p2[0]); close(p2[1]);
		
		execlp("grep", "grep", "-i", argv[2], NULL);

		cerr << "exec(grep): " << errno << endl;
		return 8;
	}

	// processo genitore
	close(p1[0]); close(p1[1]);
	
	// sostituisco lo standard input con il lato in lettura della seconda pipe
	dup2(p2[0], 0);
	// da questo momento in poi lo standard input contiene i dati provenienti dallo
	// standard output del processo gunzip
	
	close(p2[0]); close(p2[1]);
	
	execlp("more", "more", NULL);

	cerr << "exec(more): " << errno << endl;
	return 9;
}
