/***************************************************************************************
* Il programma utilizza i Thread di Windows per simulare gli studenti: ogni studente   *
*  rappresentato da un thread. I distributori automatici di biglietti sono simulati   *
* utilizzando un semaforo: l'esercizio non richiede il controllo sul singolo           *
* distributore per cui un semaforo, opportunamente dimensionato,  lo strumento pi    *
* comodo da usare. Le stesse considerazioni valgono per la rappresentazione della      *
* giostra: non  necessario sapere "quale" posto  occupato.                           *
* La gestione delle condizioni di errore non  stata implementata completamente per    *
* non appesantire la soluzione.                                                        *
****************************************************************************************/

#include <windows.h>
#include <iostream>

using namespace std;

const int MAX_WAIT_SIZE = 64; // Numero massimo handle gestito da WaitForMultipleObjects

const int TEMPO_ACQUISTO_BIGLIETTO = 20; // Secondi
const int DURATA_GIRO_SU_GIOSTRA = 350; // Secondi

const int NUM_STUDENTI = 120;
const int NUM_DISTRIBUTORI = 5;
const int CAPIENZA_GIOSTRA = 20;

// g_hDistributi  l'handle dell'oggetto Semaphore usato per simulare i distributori
// automatici di biglietti.
HANDLE g_hDistributori = NULL;

// g_hGiostra  l'handle dell'oggetto Semaphore usato per simulare la giostra.
// Il contatore interno gestito dal semaforo rappresenta i posti liberi della giostra.
HANDLE g_hGiostra = NULL;

// g_hGiostraPiena  l'handle di un oggetto Event. Serve per gestire l'evento "tutti
// i posti della giostra sono stati occupati  pu quindi iniziare a girare".
HANDLE g_hGiostraPiena = NULL;

// g_hStudentiTuttiDivertiti  l'handle di un oggetto Event usato per segnalare la
// condizione: "tutti gli studenti hanno fatto un giro in giostra".
HANDLE g_hStudentiTuttiDivertiti = NULL;

// g_hMutex  l'handle di un oggetto Mutex utilizzato per gestire l'accesso concorrente
// alle variabili globali g_nPersoneSuGiostra e g_nStudentiDaDivertire. In questo caso
//  sufficiente utilizzare un solo oggetto mutex perch le due variabili rappresentano
// una singola risorsa: lo stato della simulazione.
HANDLE g_hMutex = NULL;

int g_nPersoneSuGiostra = 0;
int g_nStudentiDaDivertire = NUM_STUDENTI;

void AcquistaBiglietto()
{
	/*	
		Attendo che un distributore sia libero per acquistare il biglietto.
		Quando WaitForSingleObject termina l'esecuzione il contatore interno
		al semaforo  decrementato di una unit.
	*/

	DWORD dwWaitResult = WaitForSingleObject(g_hDistributori, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(1);

	Sleep(TEMPO_ACQUISTO_BIGLIETTO * 1000);

	// Ho acquistato il biglietto: devo lasciare libero il distributore
	ReleaseSemaphore(g_hDistributori, 1, NULL);
}

void SaliSuGiostra()
{
	// Attendo che ci sia almeno un posto libero sulla giostra
	DWORD dwWaitResult = WaitForSingleObject(g_hGiostra, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(2);

	// Acquisisco la ownership del mutex per essere certo di essere l'unico
	// thread ad accedere alla variabile globale g_nPersoneSuGiostra
	dwWaitResult = WaitForSingleObject(g_hMutex, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(3);

	// Salgo sulla giostra
	++g_nPersoneSuGiostra;

	// Quando la giostra  piena pu iniziare il divertimento
	if(g_nPersoneSuGiostra == CAPIENZA_GIOSTRA)
		SetEvent(g_hGiostraPiena);

	// Rilascio il mutex per conterire l'accesso alla variabile g_nPersoneSuGiostra
	// ad un altro thread
	ReleaseMutex(g_hMutex);
}

void Divertiti()
{
	// Devo attendere che tutti i posti sulla giostra siano stati occupati
	DWORD dwWaitResult = WaitForSingleObject(g_hGiostraPiena, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(4);

	// Parte la giostra!
	Sleep(DURATA_GIRO_SU_GIOSTRA * 1000);
}

void ScendiGiostra()
{
	DWORD dwWaitResult = WaitForSingleObject(g_hMutex, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(5);

	--g_nPersoneSuGiostra; // scendo dalla giostra
	--g_nStudentiDaDivertire; // uno studente in pi si  divertito

	if(g_nPersoneSuGiostra == 0)
	{
		// Il turno precedente  terminato e tutti gli studenti sono scesi
		// dalla giostra. Libero quindi tutti i posti della giostra per
		// consentire l'accesso alla giostra ad altri studenti.

		ResetEvent(g_hGiostraPiena);	// la giostra non  pi piena
		ReleaseSemaphore(g_hGiostra, CAPIENZA_GIOSTRA, NULL); // Tutti i posti sono liberi
	}

	if(g_nStudentiDaDivertire == 0)
	{
		// Tutti gli studenti hanno effettuato il giro in giostra.
		SetEvent(g_hStudentiTuttiDivertiti);
	}

	ReleaseMutex(g_hMutex);
}

void AttendiCompagni()
{
	// Non posso lasciare il parco giochi da solo: devo attendere tutti i miei
	// compagni.
	DWORD dwWaitResult = WaitForSingleObject(g_hStudentiTuttiDivertiti, INFINITE);
	if(dwWaitResult != WAIT_OBJECT_0)
		ExitThread(6);
}

DWORD WINAPI Studente(LPVOID lpVoid)
{
	AcquistaBiglietto();
	SaliSuGiostra();
	Divertiti();
	ScendiGiostra();
	AttendiCompagni();

	return 0;
}

int main()
{
	HANDLE hThread[NUM_STUDENTI];
	DWORD dwWaitResult;
	int i;

	if((NUM_STUDENTI % CAPIENZA_GIOSTRA) != 0)
	{
		cerr << "Attenzione! La simulazione funziona solo se il numero degli stutendi  multiplo della capienza della giostra." << endl;
		return 10;
	}

	// Inizialmente tutti i distributori sono liberi per cui i parametri
	// lInitialCount ed lMaximumCount hanno lo stesso valore.
	g_hDistributori = CreateSemaphore(NULL, NUM_DISTRIBUTORI, NUM_DISTRIBUTORI, NULL);
	if(g_hDistributori == NULL)
		return 1;

	// Inizialmente tutti i posti sulla giostra sono liberi per cui i parametri
	// lInitialCount ed lMaximumCount hanno lo stesso valore.
	g_hGiostra = CreateSemaphore(NULL, CAPIENZA_GIOSTRA, CAPIENZA_GIOSTRA, NULL);
	if(g_hGiostra == NULL)
		return 2;

	// In questo caso gli eventi DEVONO essere manual reset altrimenti un solo 
	// thread (studente) in attesa recepisce l'evento.
	g_hGiostraPiena = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(g_hGiostraPiena == NULL)
		return 3;

	g_hStudentiTuttiDivertiti = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(g_hStudentiTuttiDivertiti == NULL)
		return 4;

	g_hMutex = CreateMutex(NULL, FALSE, NULL);
	if(g_hMutex == NULL)
		return 5;

	// Avvio i thread della simulazione
	for(i = 0; i < NUM_STUDENTI; ++i)
	{
		// La condizione di errore andrebbe gestita meglio!
		hThread[i] = CreateThread(NULL, 0, Studente, NULL, 0, NULL);
		if(hThread[i] == NULL)
			return 6;
	}

	// WaitForMultipleObjects ha un limite di 64 handle su cui attendere.
	// Per tale motivo  necessario suddividere i thread in gruppi di 64 (MAX_WAIT_SIZE).

	int CicliDiWait = NUM_STUDENTI / MAX_WAIT_SIZE;

	for(i = 0; i < CicliDiWait; ++i)
	{
		dwWaitResult = WaitForMultipleObjects(MAX_WAIT_SIZE, &hThread[i * MAX_WAIT_SIZE], TRUE, INFINITE);
		if(dwWaitResult == WAIT_FAILED)
			cerr << "WaitForMultipleObjects ha fallito. Codice di errore: " << GetLastError() << endl;
	}
	if(NUM_STUDENTI % MAX_WAIT_SIZE)
	{
		// NUM_STUDENTI non  multiplo di MAX_WAIT_SIZE
		dwWaitResult = WaitForMultipleObjects(NUM_STUDENTI % MAX_WAIT_SIZE, &hThread[CicliDiWait * MAX_WAIT_SIZE], TRUE, INFINITE);
		if(dwWaitResult == WAIT_FAILED)
			cerr << "WaitForMultipleObjects ha fallito. Codice di errore: " << GetLastError() << endl;
	}

	// Terminata la simulazione bisogna chiudere tutti gli handle
	for(i = 0; i < NUM_STUDENTI; ++i)
		CloseHandle(hThread[i]);

	CloseHandle(g_hMutex);
	CloseHandle(g_hGiostra);
	CloseHandle(g_hGiostraPiena);
	CloseHandle(g_hStudentiTuttiDivertiti);
	CloseHandle(g_hDistributori);

	return 0;
}
